#!/usr/bin/ruby

# Synchronous Azure actions: password changing and user deletion

# TODO: Combine HTTP response handling. Dedupe 50 lines of code.

require_relative './synchronous_action'

require 'adal'
require 'http'
require 'yaml'
require 'securerandom'


MS_GRAPH_API_URL = 'https://graph.microsoft.com'.freeze


class SynchronousAzureAction < SynchronousAction::SynchronousAction
  def handle_post(request, response)
    # Exceptions generated inside this method will be handled and
    # logged by the parent class

    if request.path == '/azure_delete_user' && @enabled.include?(:delete_user)
      delete_user(request, response)
    elsif request.path == '/azure_change_password' && @enabled.include?(:change_password)
      change_password(request, response)
    else
      response.status = 404
      response.body = ''
    end
  end

  def system_name
    'azure'
  end

  def application_status_codes
    Set.new(%I[invalid_username user_not_found unmet_password_requirements]).freeze
  end

  private

  # Logs in to Azure and acquires an access token that is used to
  # authenticate HTTP requests
  def acquire_access_token
    begin
      # Thank you Microsoft for making this so ridiculously complicated.
      # Just let me customize the logger output myself.

      # I spent over two hours trying to override this logging system. I just
      # need the message itself. Everything else (severity, timestamps, etc.)
      # is just noise. But nah, you can't do that. They say you can monkeypatch
      # everything in Ruby. No, no you can't. Not this you can.

      # I want LESS, not more. LESS INFORMATION. Because I already have that
      # information. I just want the MESSAGE. Nothing more.

      # I fear the day when multiple logins happen simultaneously. The logger
      # in ADAL is a global object, so from that point on everything is broken.

      random_filename = "/tmp/#{SecureRandom.hex(32)}-#{@conf['azure']['domain']}.log"
      output = File.open(random_filename, 'w')
      ADAL::Logging.log_output = output

      # Authenticate using ADAL
      adal_token_cache = ADAL::MemoryCache.new

      user_cred = ADAL::UserCredential.new(
        @conf['azure']['username'],
        @conf['azure']['password']
      )

      client_cred = ADAL::ClientCredential.new(
        @conf['azure']['client_id'],
        @conf['azure']['client_secret']
      )

      context = ADAL::AuthenticationContext.new(
        ADAL::Authority::WORLD_WIDE_AUTHORITY,
        @conf['azure']['tenant_id'],
        :token_cache => adal_token_cache
      )

      adal_token = context.acquire_token_for_user(
        MS_GRAPH_API_URL,
        client_cred,
        user_cred
      )

      if adal_token.error?
        output.close

        # just... give... me... the... messages...
        messages = File.read(random_filename)
        File.delete(random_filename)

        # remove garbage \r newlines
        messages = messages.delete("\r").split("\n")

        messages.each do |l|
          log_error(l)
        end

        return nil
      end

      return adal_token.access_token
    rescue StandardError => e
      log_error(e)
      return nil
    end
  end

  # ------------------------------------------------------------------------------------------------
  # Delete a user from Azure

  def delete_user(request, response)
    return unless validate_common_request_params(request, response, 'delete_user')

    unless all_params?('organisation', 'user', 'user_id', 'school', 'school_id')
      log_error("delete_user(): Missing a required field in the \"parameters\" block, raw data:")
      log_error("  #{@parameters.inspect}")

      return fail(response, :incomplete_request)
    end

    unless @conf.include?('azure')
      log_error('delete_user(): Missing "azure" section in the configuration')
      return fail(response, :configuration_error)
    end

    organisation = @parameters[0]['organisation']

    # Domain matching
    if organisation != @conf['azure']['incoming_puavo_domain']
      log_error("delete_user(): Request has invalid Puavo domain \"#{organisation}\"")
      return fail(response, :configuration_error)
    end

    log_info('delete_user(): Attempting AD login')

    access_token = acquire_access_token()

    if access_token.nil?
      log_error('delete_user(): Unable to acquire an access token, aborting')
      return fail(response, :bad_credentials)
    end

    username = @parameters[0]['user']

    full_username = "#{username}@#{@conf['azure']['domain']}"

    log_info("delete_user(): Deleting user \"#{username}\", " \
             "full username is \"#{full_username}\"")

    begin
      # TODO: Replace this with Ruby's net/http to remove a gem dependency
      http = HTTP.timeout(connect: 5, read: 10)
                 .auth("Bearer #{access_token}")
                 .headers('Content-type' => 'application/json')

      res = http.delete("https://graph.microsoft.com/v1.0/users/#{full_username}")
      #res = http.get("https://graph.microsoft.com/v1.0/users/#{full_username}")
    rescue StandardError => e
      log_error('delete_user(): Network request failed:')
      log_error("delete_user():   #{e}")
      return fail(response, :network_error)
    end

    if res.code == 200 || res.code == 204   # 204 = "no content", assume it's good
      log_info('delete_user(): User deletion complete')
      return success(response, :ok)
    end

    # Interpret results
    unless res.body.to_s.empty?
      begin
        data = JSON.parse(res.body.to_s)
      rescue StandardError
        log_error('delete_user(): Unable to parse the server response:')
        log_error("delete_user():   \"#{res.body.to_s}\"")
        return fail(response, :malformed_reply)
      end

      code = data['error']['code']
      message = data['error']['message']
    else
      code = ''
      message = ''
    end

    log_error('delete_user(): Received an error reply:')
    log_error("   code: \"#{code}\"")
    log_error("   message: \"#{message}\"")

    if res.code == 400    # bad request
      if message.include?('password does not comply with password complexity')
        log_error('delete_user(): The password does not meet the complexity requirements')
        return fail(response, :unmet_password_requirements)
      end

      log_error('Bad request')
      return fail(response, :bad_request)
    end

    if code == 'Authorization_RequestDenied'
      log_error('delete_user(): Unauthorized')
      return fail(response, :unauthorized)
    elsif code == 'Request_ResourceNotFound'
      log_error('delete_user(): The requested user was not found')
      return fail(response, :user_not_found)
    end

    log_error('delete_user(): Unknown error')
    fail(response, :unknown_error)
  end

  # ------------------------------------------------------------------------------------------------
  # Synchronise password changes from Puavo to Azure

  def change_password(request, response)
    return unless validate_common_request_params(request, response, 'change_password')

    unless all_params?('organisation', 'user', 'new_password')
      log_error('change_password(): Missing a required field in the "parameters" block, raw data:')
      log_error("  #{@parameters.inspect}")

      return fail(response, :incomplete_request)
    end

    unless @conf.include?('azure')
      log_error('change_password(): Missing "azure" section in the configuration')
      return fail(response, :configuration_error)
    end

    organisation = @parameters[0]['organisation']
    username = @parameters[0]['user']
    new_password = @parameters[0]['new_password']

    if username.nil? || username.empty? || new_password.nil? || new_password.empty?
      log_error('change_password(): Missing or empty username/password')
      return fail(response, :incomplete_request)
    end

    # No admin user password changes
    if username.match(/^adm/) || username.match(/^admin/)
      log_error("change_password(): Refusing to sync password for admin user \"#{username}\" " \
                "in organisation \"#{@parameters[0]['organisation']}\"")

      # set custom HTTP code
      return set_reply(response, false, :invalid_username, nil, 403)
    end

    # Domain matching
    if organisation != @conf['azure']['incoming_puavo_domain']
      log_error("change_password(): Request has invalid Puavo domain \"#{organisation}\"")
      return fail(response, :configuration_error)
    end

    log_info('change_password(): Attempting AD login')

    access_token = acquire_access_token()

    if access_token.nil?
      log_error('change_password(): Unable to acquire an access token, aborting')
      return fail(response, :bad_credentials)
    end

    full_username = "#{username}@#{@conf['azure']['domain']}"

    log_info("change_password(): Changing the password for user " \
             "\"#{username}\", full username is \"#{full_username}\"")

    # Change the password
    request = {
      passwordProfile: {
        password: new_password,
        forceChangePasswordNextSignIn: false
      },

      passwordPolicies: 'DisableStrongPassword'
    }

    begin
      # TODO: Replace this with Ruby's net/http to remove a gem dependency
      http = HTTP
        .timeout(connect: 5, read: 10)
        .auth("Bearer #{access_token}")
        .headers('Content-type' => 'application/json')

      res = http.patch(
        "https://graph.microsoft.com/v1.0/users/#{full_username}",
        :json => request
      )
    rescue StandardError => e
      log_error('change_password(): Network request failed:')
      log_error("change_password():   #{e}")
      return fail(response, :network_error)
    end

    if res.code == 200 || res.code == 204   # 204 = "no content", assume it's good
      log_info('change_password(): Password change complete')
      return success(response, :ok)
    end

    # Interpret results
    unless res.body.to_s.empty?
      begin
        data = JSON.parse(res.body.to_s)
      rescue StandardError
        log_error('change_password(): Unable to parse the server response:')
        log_error("change_password():   \"#{res.body.to_s}\"")
        return fail(response, :malformed_reply)
      end

      code = data['error']['code']
      message = data['error']['message']
    else
      code = ''
      message = ''
    end

    log_error('change_password(): Received an error reply:')
    log_error("   code: \"#{code}\"")
    log_error("   message: \"#{message}\"")

    if res.code == 400    # bad request
      if message.include?('password does not comply with password complexity')
        log_error('change_password(): The password does not meet the complexity requirements')
        return fail(response, :unmet_password_requirements)
      end

      log_error('Bad request')
      return fail(response, :bad_request)
    end

    if code == 'Authorization_RequestDenied'
      log_error('change_password(): Unauthorized')
      return fail(response, :unauthorized)
    elsif code == 'Request_ResourceNotFound'
      log_error('change_password(): The requested user was not found')
      return fail(response, :user_not_found)
    end

    log_error('change_password(): Unknown error')
    fail(response, :unknown_error)
  end
end

# --------------------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------------------

SynchronousAction::launch(SynchronousAzureAction)
