#!/usr/bin/ruby

# Synchronous Google actions: password changing and user deletion

require_relative './synchronous_action'

require 'googleauth'
require 'googleauth/stores/file_token_store'
require 'google/apis/admin_directory_v1'

class GoogleError < StandardError; end
class NotFound < StandardError; end
class ReusedPassword < StandardError; end

class SynchronousGoogleAction < SynchronousAction::SynchronousAction
  def handle_post(request, response)
    # Exceptions generated inside this method will be handled and
    # logged by the parent class

    if request.path == '/google_delete_user' && @enabled.include?(:delete_user)
      delete_user(request, response)
    elsif request.path == '/google_change_password' && @enabled.include?(:change_password)
      change_password(request, response)
    else
      response.status = 404
      response.body = ''
    end
  end

  def system_name
    'google'
  end

  def application_status_codes
    Set.new(%I[invalid_username user_not_found unmet_password_requirements reused_password]).freeze
  end

  private

  # ------------------------------------------------------------------------------------------------
  # Google API stuff

  # Logs in to Google API. Only the new service account -style logins with a private key
  # are supported. Google does not even support the old OAuth2 style logins anymore.
  # See https://developers.google.com/identity/protocols/oauth2/service-account.
  # (September/October 2022.)
  def login_to_google(settings)
    # https://github.com/googleapis/google-auth-library-ruby#example-service-account
    credentials = Google::Auth::ServiceAccountCredentials.make_creds(
      json_key_io: File.open(File.join(settings['directory'], 'key.json')),
      scope: [Google::Apis::AdminDirectoryV1::AUTH_ADMIN_DIRECTORY_USER]
    )

    credentials.sub = settings['delegate']    # undocumented in practice, but required
    credentials.fetch_access_token!

    # http://www.rubydoc.info/github/google/google-api-ruby-client/Google/Apis/AdminDirectoryV1/DirectoryService
    client = Google::Apis::AdminDirectoryV1::DirectoryService.new

    # http://www.rubydoc.info/github/google/google-api-ruby-client/Google/Apis/ClientOptions
    client.client_options.application_name = 'google_sync_actions'

    client.authorization = credentials

    client
  end

  # Performs a Google API operation. Handles retries, backoff timing, and many other things.
  def do_google_operation(&block)
    attempt_num = 1
    wait_time = 1

    begin
      block.call
    rescue Google::Apis::ServerError => e
      # Server errors can be retried
      log_error('do_google_operation(): Server error!')
      log_error("do_google_operation():   #{e}")

      attempt_num += 1

      if attempt_num > MAX_ATTEMPTS
        log_error('do_google_operation(): Retry count reached, stopping')
        raise GoogleError
      end

      # See "Implementing Exponential Backoff" in
      # https://developers.google.com/admin-sdk/directory/v1/limits
      time = wait_time + rand(0.1..1.0)
      logger.info("do_google_operation(): Attempt number #{attempt_num} in #{time.round(1)} seconds...]")
      sleep(time)
      wait_time *= 2

      retry
    rescue Google::Apis::ClientError => e
      # Client errors cannot be retried
      log_error('do_google_operation(): Client error!')
      reason = JSON.parse(e.body)['error']['errors'][0]['reason']

      if reason == 'duplicate'
        # A user with this name already exists, but for some reason we're not detecting it
        log_error('do_google_operation(): Google says this is a duplicate entry?')
      elsif reason == 'quotaExceeded'
        # Can happen, for example, if you try to create too many organisation units too quickly.
        # If we can, we'll wait for a few seconds and try again.
        log_error('do_google_operation(): Quota exceeded')

        attempt_num += 1

        if attempt_num > MAX_ATTEMPTS
          log_error('do_google_operation(): Retry count reached, stopping')
        else
          # See "Implementing Exponential Backoff" in
          # https://developers.google.com/admin-sdk/directory/v1/limits
          time = wait_time + rand(0.1..1.0)
          log_error("do_google_operation(): Attempt number #{attempt_num} in #{time.round(1)} seconds...]")
          sleep(time)
          wait_time *= 2

          retry
        end
      elsif reason == 'limitExceeded'
        # Happens, for example, if you try to add more users to a domain than
        # what it is currently configured for
        log_error('do_google_operation(): Domain limit exceeded!')
        log_error("do_google_operation():   #{e}")
        log_error("do_google_operation():   #{e.body}")
        log_error('do_google_operation(): Cannot continue, stopping here!')
      elsif reason == 'notFound'
        # technically, this is not an error
        log_error('do_google_operation(): The requested user does not exist')
        raise NotFound
      elsif e.to_s.include?('Reused Password')
        log_error('do_google_operation(): Reused password')
        log_error("do_google_operation():   #{e}")
        log_error("do_google_operation():   #{e.body}")
        log_error('do_google_operation(): Cannot continue, stopping here!')
        raise ReusedPassword
      else
        log_error('do_google_operation(): Unhandled error, see the raw dump below')
        log_error("do_google_operation():   #{e}")
        log_error("do_google_operation():   #{e.body}")
        log_error('do_google_operation(): Cannot retry, stopping here!')
      end

      raise GoogleError
    rescue Google::Apis::AuthorizationError => e
      # This should not happen, but...
      log_error('do_google_operation(): Authorisation error!')
      log_error("do_google_operation():   #{e}")
      log_error('do_google_operation(): Cannot retry, stopping here!')
      raise GoogleError
    rescue Google::Apis::RateLimitError => e
      # This one's so bad that we'll just stop right here
      log_error('do_google_operation(): Rate limit error!')
      log_error("do_google_operation():   #{e}")
      log_error('do_google_operation(): Cannot retry, stopping here!')
      raise GoogleError
    end
  end

  # ------------------------------------------------------------------------------------------------
  # Delete a user from Google

  def delete_user(request, response)
    return unless validate_common_request_params(request, response, 'delete_user')

    unless all_params?('organisation', 'user', 'user_id', 'school', 'school_id')
      log_error('delete_user(): Missing a required field in the "parameters" block, raw data:')
      log_error("  #{@parameters.inspect}")

      return fail(response, :incomplete_request)
    end

    unless @conf.include?('google')
      log_error('delete_user(): Missing "google" section in the configuration')
      return fail(response, :configuration_error)
    end

    unless @conf['google'].include?('directory')
      log_error('delete_user(): Missing "directory" in the "google" section')
      return fail(response, :configuration_error)
    end

    unless @conf['google'].include?('domain')
      log_error('delete_user(): Missing "domain" in the "google" section')
      return fail(response, :configuration_error)
    end

    unless Dir.exist?(@conf['google']['directory'])
      log_error("delete_user(): Directory \"#{@conf['google']['directory']}\" does not exist")
      return fail(response, :configuration_error)
    end

    # Login to Google
    log_info('delete_user(): Attempting Google login')

    begin
      client = login_to_google(@conf['google'])
    rescue StandardError => e
      log_error('delete_user(): Unable to login to Google:')
      log_error("delete_user():   #{e}")
      return fail(response, :server_error)
    end

    # Retrieve the user
    full_address = "#{@parameters[0]['user']}@#{@conf['google']['domain']}"
    log_info("delete_user(): Google login complete, looking for user \"#{@parameters[0]['user']}\"" \
             " (\"#{full_address}\")")

    user = nil

    begin
      do_google_operation do
        user = client.get_user(full_address)
      end
    rescue GoogleError
      log_error('delete_user(): Unable to retrieve the user from Google')
      return fail(response, :server_error)
    rescue NotFound
      # It's up to the requester to decide if this is a fatal error or not
      return fail(response, :user_not_found)
    rescue StandardError => e
      log_error('delete_user(): Unable to retrieve the user from Google:')
      log_error("delete_user():   #{e}")
      return fail(response, :unknown_error)
    end

    # Delete the user
    log_info("delete_user(): Found user \"#{user.id}\", deleting")

    begin
      do_google_operation do
        client.delete_user(user.id)
      end
    rescue GoogleError
      log_error('delete_user(): Unable to delete the user')
      return fail(response, :server_error)
    rescue StandardError => e
      log_error('delete_user(): Unable to delete the user:')
      log_error("delete_user():   #{e}")
      return fail(response, :unknown_error)
    end

    # If we get here, the deletion succeeded
    log_info('delete_user(): User deletion complete')
    success(response, :ok)
  end

  # ------------------------------------------------------------------------------------------------
  # Synchronise password changes from Puavo to Google

  def change_password(request, response)
    return unless validate_common_request_params(request, response, 'change_password')

    unless all_params?('organisation', 'user', 'new_password')
      log_error('change_password(): Missing a required field in the "parameters" block, raw data:')
      log_error("  #{@parameters.inspect}")

      return fail(response, :incomplete_request)
    end

    unless @conf.include?('google')
      log_error('change_password(): Missing "google" section in the configuration')
      return fail(response, :configuration_error)
    end

    unless @conf['google'].include?('directory')
      log_error('change_password(): Missing "directory" in the "google" section')
      return fail(response, :configuration_error)
    end

    unless @conf['google'].include?('domain')
      log_error('change_password(): Missing "domain" in the "google" section')
      return fail(response, :configuration_error)
    end

    unless Dir.exist?(@conf['google']['directory'])
      log_error("change_password(): Directory \"#{@conf['google']['directory']}\" does not exist")
      return fail(response, :configuration_error)
    end

    username = @parameters[0]['user']
    new_password = @parameters[0]['new_password']

    if username.nil? || username.empty? || new_password.nil? || new_password.empty?
      log_error('change_password(): Missing or empty username/password')
      return fail(response, :incomplete_request)
    end

    # No admin user password changes
    if username.match(/^adm/) || username.match(/^admin/)
      log_error("change_password(): Refusing to sync password for admin user \"#{username}\" " \
                "in organisation \"#{@parameters[0]['organisation']}\"")

      # set custom HTTP code
      return set_reply(response, false, :invalid_username, nil, 403)
    end

    # Validate the password
    if new_password.length < 8
      log_error('change_password(): The password is shorter than 8 characters')
      return fail(response, :unmet_password_requirements)
    end

    # why Google, why?
    unless new_password.ascii_only?
      log_error('change_password(): The password contains non-ASCII characters')
      return fail(response, :unmet_password_requirements)
    end

    # I just don't get it...
    if new_password[0] == ' ' || new_password[-1] == ' '
      log_error('change_password(): The password starts or ends in a whitespace')
      return fail(response, :unmet_password_requirements)
    end

    # Login to Google
    log_info('change_password(): Attempting Google login')

    begin
      client = login_to_google(@conf['google'])
    rescue StandardError => e
      log_error('change_password(): Unable to login to Google:')
      log_error("change_password():   #{e}")
      return fail(response, :server_error)
    end

    # Retrieve the user
    full_address = "#{@parameters[0]['user']}@#{@conf['google']['domain']}"
    log_info('change_password(): Google login complete, looking for user ' \
             "\"#{@parameters[0]['user']}\" (\"#{full_address}\")")

    user = nil

    begin
      do_google_operation do
        user = client.get_user(full_address)
      end
    rescue GoogleError
      log_error('change_password(): Unable to retrieve the user from Google')
      return fail(response, :server_error)
    rescue NotFound
      # It's up to the requester to decide if this is a fatal error or not
      return fail(response, :user_not_found)
    rescue ReusedPassword
      log_error('change_password(): Reused password rejected')
      return fail(response, :reused_password)
    rescue StandardError => e
      log_error('change_password(): Unable to retrieve the user from Google:')
      log_error("change_password():   #{e}")
      return fail(response, :unknown_error)
    end

    # Change the password
    log_info("change_password(): Found user \"#{user.id}\", changing the password")

    begin
      do_google_operation do
        user.password = new_password
        client.patch_user(user.id, user)
      end
    rescue GoogleError
      log_error('change_password(): Unable to change the password')
      return fail(response, :server_error)
    rescue StandardError => e
      log_error('change_password(): Unable to change the password:')
      log_error("change_password():   #{e}")
      return fail(response, :unknown_error)
    end

    # If we get here, the password change succeeded
    log_info('change_password(): Password change complete')
    success(response, :ok)
  end
end

# --------------------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------------------

SynchronousAction::launch(SynchronousGoogleAction)
