require 'webrick'
require 'set'
require 'json'
require 'yaml'

module SynchronousAction
  # Default IP whitelist, used if no whitelist is specified for the service
  BUILTIN_IP_WHITELIST = Set.new(
    [
      '10.246.84.21', '10.246.84.22', '10.246.84.23',   # puavo-rest
      '10.246.84.41', '10.246.84.42', '10.246.84.43',   # puavo-web
    ]
  ).freeze

  # Built-in status codes
  BUILTIN_STATUS_CODES = Set.new(
    %I[
      ok
      unknown_version
      unknown_system
      unknown_operation
      configuration_error
      bad_request
      unauthorized
      bad_credentials
      remote_server_did_not_respond
      remote_server_refused
      incomplete_request
      malformed_reply
      server_error
      network_error
      unknown_error
    ]
  ).freeze

  class RequestParams
    attr_accessor :version, :request_id, :request_timestamp, :operation, :system

    def initialize(json = nil)
      if json
        @version = json.fetch('version', nil)
        @request_id = json.fetch('request_id', nil)
        @request_timestamp = json.fetch('request_timestamp', nil)
        @operation = json.fetch('operation', nil)
        @system = json.fetch('system', nil)
      else
        @version = nil
        @request_id = nil
        @request_timestamp = nil
        @operation = nil
        @system = nil
      end
    end

    def valid?
      return false if @version.nil?
      return false if @request_id.nil?
      return false if @request_timestamp.nil?
      return false if @operation.nil?
      return false if @system.nil?

      true
    end
  end

  class SynchronousAction < WEBrick::HTTPServlet::AbstractServlet
    def initialize(server, config)
      super(server)

      # "@config" already exists in WEBrick
      @conf = config

      @request = nil
      @parameters = nil

      # which services are enabled?
      @enabled = Set.new(config['enable'].collect(&:to_sym))
    end

    def do_POST(request, response)
      begin
        handle_post(request, response)
      rescue StandardError => e
        if @request
          id = @request.request_id

          @logger.error("[#{id}] INTERNAL SERVER ERROR!")
          @logger.error("[#{id}] #{e}")
          @logger.error("[#{id}] #{e.backtrace.join("\n")}")
        else
          @logger.error('INTERNAL SERVER ERROR!')
          @logger.error(e)
          @logger.error(e.backtrace.join("\n"))
        end

        response.status = 500
        response.body = ''
      end
    end

    # explicitly reject all GETs
    def do_GET(_, response)
      response.status = 404
    end

    # You want to override this in your class. Return false to trigger a 404.
    def handle_post(request, response)
      false
    end

    # You also want to override this, to provide a system name
    def system_name
      'you_forgot_to_override_this_method'
    end

    # Override this to provide valid custom status codes
    def application_status_codes
      Set.new([]).freeze
    end

    def generate_call_id
      'ABCDEGIJKLMOQRUWXYZ12346789'.split('').sample(10).join
    end

    protected

    def set_raw_error(response, status)
      response.status = status
      response.body = ''
      false
    end

    def set_reply(response, success, code, message = nil, status_code = 200)
      if !BUILTIN_STATUS_CODES.include?(code) && !application_status_codes.include?(code)
        # you just had to screw this up, didn't you?
        log_error("Unknown reply code \"#{code}\"! Overriding reply.")

        success = false
        code = :server_error
        message = 'SynchronousAction::set_reply(): invalid message code'
      end

      reply = {
        'version': 0,       # currently always 0
        'operation': @request.operation,
        'system': @request.system,
        'request_id': @request.request_id,
        'request_timestamp': @request.request_timestamp,
        'reply_id': generate_call_id,
        'reply_timestamp': Time.now.utc.to_i,
        'success': success,
        'code': code.to_s
      }

      reply['message'] = message if message

      log_info("Sending back reply \"#{reply[:reply_id]}\", success=#{success}, code=\"#{code}\", " \
               "message=\"#{message}\", status_code=#{status_code}")

      response.status = status_code
      response.body = reply.to_json
      response['Content-Type'] = 'text/json'

      success
    end

    def success(response, code, message = nil)
      set_reply(response, true, code, message)
    end

    def fail(response, code, message = nil)
      set_reply(response, false, code, message)
    end

    # do not call this unless @request has been set up!
    def log_info(message)
      @logger.info("[#{@request.request_id}] #{message}")
    end

    # do not call this unless @request has been set up!
    def log_error(message)
      @logger.error("[#{@request.request_id}] #{message}")
    end

    # Checks that all the listed keys exists in @parameters. Returns false if even one is missing.
    def all_params?(*names)
      names.each do |name|
        @parameters.each do |params|
          unless params.include?(name)
            @logger.warn("[#{@request.request_id}] Missing parameter \"#{name}\" in \"#{params}\"")
            return false
          end
        end
      end

      true
    end

    def validate_common_request_params(request, response, operation_name)
      ip = request.remote_ip

      # IP check
      ip_whitelist = \
        if @conf['service'].include?('ip_whitelist')
          Set.new(@conf['service']['ip_whitelist'])
        else
          BUILTIN_IP_WHITELIST
        end

      unless ip_whitelist.include?(ip)
        @logger.error("Rejecting a request from a non-whitelisted address \"#{ip}\"")
        return set_raw_error(response, 403)
      end

      body = request.body

      if body.nil? || body.empty?
        @logger.error("Received a broken request from \"#{ip}\". Missing POST body data.")
        return set_raw_error(response, 400)
      end

      begin
        data = JSON.parse(body)
      rescue StandardError => e
        @logger.error("Received a broken request from \"#{ip}\". Unable to parse JSON data:")
        @logger.error(e.to_s)
        @logger.error('Raw request body data:')
        @logger.error(body.to_s)
        return set_raw_error(response, 400)
      end

      # Validate the required parameters
      @request = RequestParams.new(data)

      unless @request.valid?
        @logger.error("Received a request with one or more missing fields from \"#{ip}\": #{data}")
        return set_raw_error(response, 400)
      end

      # log_error() and log_info() can be used from this point on, as @request has been set up

      log_info("operation=\"#{@request.operation}\", \"system=\"#{@request.system}\", sent from " \
               "\"#{ip}\" at #{Time.at(@request.request_timestamp).strftime('%Y-%m-%d %H:%M:%S')}")

      # Validate the request parameters. If the version number ever changes,
      # this check must be done earlier, as different versions can have
      # different parameters in the request.
      if @request.version != 0
        log_error("Invalid version \"#{@request.version}\", request data:")
        log_error("  #{data}")
        return fail(response, :invalid_version)
      end

      if @request.system != system_name
        log_error("Unknown system \"#{@request.system}\", request data:")
        log_error("  #{data}")
        return fail(response, :unknown_system)
      end

      if @request.operation != operation_name
        log_error("Unknown operation \"#{@request.operation}\", request data:")
        log_error("  #{data}")
        return fail(response, :unknown_operation)
      end

      # this block is always required, even if it's empty
      unless body.include?('parameters')
        log_error('Missing "parameters" in request, request data:')
        log_error("  #{data}")
        return fail(response, :incomplete_request)
      end

      # make the request parameters accessible
      if data['parameters'].instance_of?(Hash)
        @parameters = [data['parameters'].dup]
      elsif data['parameters'].instance_of?(Array)
        @parameters = data['parameters'].dup
      else
        log_error("Expected a Hash or Array 'parameters', but got #{clazz} instead:")
        log_error("  #{data['parameters']}")
        return fail(response, :bad_request)
      end

      if @parameters.empty?
        log_error('Did not receive any requests')
        return fail(response, :bad_request)
      end

      log_info("Have #{@parameters.count} request(s) in this batch")

      true
    end
  end

  # Run the service
  def self.launch(class_instance_name)
    # Systemd wants "simple" command lines, so pass the configuration file name in environment
    # variables. Otherwise we'd have to create a shell script.
    unless ENV.include?('CONFIG_FILE')
      puts "ERROR: No 'CONFIG_FILE' in environment variables"
      puts "Please run the script this way:\n\n"
      puts "\tCONFIG_FILE=\"/full/path/to/config/file.yml\" #{$PROGRAM_NAME}\n\n"
      exit(1)
    end

    # Parse the configuration file
    config_file = ENV['CONFIG_FILE']

    begin
      config = YAML.safe_load(File.open(config_file).read)
    rescue StandardError => e
      puts "ERROR: Can't load configuration file \"#{config_file}\": #{e}"
      exit(1)
    end

    # Launch the server
    puts "Using configuration file \"#{config_file}\""

    server = WEBrick::HTTPServer.new(Port: config['service']['port'])
    server.mount '/', class_instance_name, config
    trap('INT') { server.shutdown }
    server.start
  end
end
